! Program: TIMEFRQ 4.4
!
! Purpose: Time-frequency analysis of unevenly spaced time series.
! --------
!
! Description: For a frequency (filtfrq) in [fstart, fstop] the time-dependent
! ------------ amplitude of a sinusoidal regression curve is estimated using a
!         sliding rectangular window. The window is shifted consecutively
!         by one data point along the time axis of the input time series.
!         The window width and, hence, the length of the segments on which
!         the amlitude estimates are based is:
!
!                     Tseg = wfac * Tper
!
!         where Tseg is the length of the segments, Wfac is a width-factor
!         and Tper = 1/filtfrq is the periodicity of interest. Note, that
!         in case of unevenly spaced time series, the segments are of "equal"
!         duration but do not necessarily contain the same number of data
!         points.
!
!         The half-amplitude bandwidth of the filter is approximateley
!
!                       BW = beta /(Wfac * TPer)
!
!         where beta depends on the window as follows:
!
!         Window             beta
!         -----------------------
!         Rectangular        1.14
!         Welch              1.54
!         Hanning            2.09
!         Triangular         1.70
!         Blackmann Harris   2.38
!
!
!         The estimated amplitudes and phases are saved vs.the average time
!         of the segment.

!
!         For each filtfrq, estimated time-dependent-amplitude as well as the
!         resulting time vector are saved to file FNOUT.
!
! Input: - standard time series files (may be unevenly spaced)
! ------ - paramter input via namelist:
!
!           &cfg
!              fnin = 'env_test.dat',    -> input file
!             fnout = 'env_test.tf',     -> output: time, freq. and amplitude
!            fstart = 0.005,             -> lowest frequency to analyze
!             fstop = 0.02,              -> lowest frequency to analyze
!                df = 0.0001,            -> freq. spacing in [fstart, fstop]
!              wfac = 4.0,               -> relative width of sliding window
!             dtout = 0.0,               -> time-interval between starting time
!                                           of sliding window; if set to 0.0
!                                           window is shifted by one data point
!                                           [default = 0.0]
!               iwin = 2                 -> window type (0: Rectangular [default]
!                                                        1: Welch
!                                                        2: Hanning
!                                                        3: Triangular
!                                                        4: Blackman-Harris)
!            bwtest = .true.             -> if true, frequency response function
!                                           is determined [default = false]
!              gplt = .true.             -> if true blocks of separate frequencies in
!                                           output are separated by blank, i.e., gnuplot
!                                           compatible [default = false]
!
! Output: FNOUT
! ------- bw.plt      (frequency response of the filter; Gnuplot script)
!         Nyq_Err.dat
!         TimeSeries.avg
!
! Ref.: Ferraz-Mello, 1981: Estimation of periods from unequally
! ----- spaced observations, Astron. J., 86, 619-624.
!
!
! written        : 09.03.99 (adapted from ENVELOPE4)
! latest revision: 29.02.00 V. 2.0
!                           - option for tapering segments for better
!                             sidelobe suppression (new option IWin in
!                             cfg file)
!                           - subtract mean prior to tapering (time series
!                             with negative mean caused problems if tapers
!                             other than rectangular were used)
!                  08.05.00 - V. 3.0
!                           - new config parameter DTOUT (useful for large
!                             data files)
!                  24.06.01 - V. 4.0
!                           - output changed; t,f,ampl. are directly saved to
!                             ASCII file
!                           - delta freq. (df) is read from namelist instead of
!                             number of freqs. (nfrq)
!                           - times where Nyq. criterion is violated are
!                             written to file Nyq_Err.dat
!                  06.11.02 - V. 4.1
!                           - write local Nyquist freq to Nyq_Err.dat
!                  10.07.04 - V. 4.2
!                           - option for gnuplot output added
!                  19.7.04  - V. 4.3
!                           - option for (automatic) check of input TS added
!                             -> program stops upon descending time axis
!                             -> data for mult. sampling times are averaged
!                                (averaged time series is written to TimeSeries.avg)
!                  24.05.04 - length of file name incr. to 80
!                  20.08.18 - V. 4.4
!                           - bugfix: prevent rare runtime errors when working on the
!                                     last segment (thanks to Manfred Mudelsee)
!
! Author:  Michael Schulz
! -------  MARUM and Dept. for Geosciences, Univ. Bremen
!          mschulz@marum.de
!          https://www.marum.de/en/Michael-Schulz.html
!
!--------------------------------------------------------------------------
!		     M O D U L E   D E F I N I T O N S
!--------------------------------------------------------------------------
! global array dimensions and constants
! -------------------------------------
  module const
     implicit none
     public
     integer, parameter :: maxdim = 25000
     real, parameter    :: dumval = -9999.
     real, parameter    :: pi = 3.141592654
     real, parameter    :: twopi = 2.0 * pi
     character(len = *), parameter :: prgvers = 'TIMEFRQ 4.4'
  end module const
!
! time series data
! ----------------
  module timeser
     use const
     implicit none
     public
     real, dimension(maxdim) :: t, x
     integer :: np
  end module timeser
!
!--------------------------------------------------------------------------
!                 E N D   M O D U L E   D E F I N I T O N S
!--------------------------------------------------------------------------
  program timefrq
!
  use const
  use timeser
  use mutil
!
  implicit none
!
  real, dimension(maxdim) :: tout, ampl
  real, dimension(:, :), allocatable :: astor, tstor
  real    :: fstart, fstop, filtfrq, wfac, tseg, df
  real    :: f0, fnow, arg, dt, fnyq
  integer :: nsegout, i, j, iocheck, ncnt, nfrq
  real    :: dtout = 0.0
  integer :: iwin = 0
  integer :: nyqerr = 0
  logical :: bwtest = .false.
  logical :: gplt = .false.
  logical :: err = .false.
  character (len = 80) :: fnin, fnout, cfgfile
!
  namelist /cfg/ fnin, fnout, wfac, fstart, fstop, df, bwtest, iwin, dtout, gplt
!
! check number of command line arguments
! --------------------------------------
  ncnt = nargs()
  if (ncnt .ne. 1) then
     write (*, '(//1x,a)') 'Error - pass config file via the command line!'
     stop ' '
  end if
!
! retrieve command line arguments
! -------------------------------
  cfgfile = getarg(1)
  open (10, file = cfgfile, form = 'formatted', status = 'old', &
       iostat = iocheck)
  if (iocheck .ne. 0 ) then
     write (*, '(//a)') ' Error - Can''t open config file!'
     stop ' '
  end if
  read(10, nml = cfg)
  close (10)
!
! retrieve time series data
! -------------------------
  call readdat(fnin, t, x, np)
!
! check input time axis
! ---------------------
  call check(t, x, np)
!  do i = 1, np
!     print*,t(i),x(i)
!  end do
! stop
!
! check if fmax is less than local Nyquist frequency
! --------------------------------------------------
  open (30, file = "Nyq_Err.dat")
  write(30,'(a27)') '# Times where fstop > fnyq:'
  do i = 2, np
     dt = t(i) - t(i-1)
     fnyq = 0.5 / dt
     if (fstop .gt. fnyq) then
        err = .true.
        nyqerr = nyqerr + 1
        write(30,'(2(e12.6,3x))') t(i), fnyq
     end if
  end do
  if (err .eqv. .true.) then
     close (30, status = 'keep')
  else
     close (30, status = 'delete')
  end if
!
! determine number of frequencies
! -------------------------------
  nfrq = nint((fstop - fstart) / df) + 1
!
! set up work space
! -----------------
  allocate (astor(np, nfrq))
  allocate (tstor(np, nfrq))
  astor(:,:) = dumval
  tstor(:,:) = dumval
!
! loop over frequencies
! ---------------------
  do i = 1, nfrq
     filtfrq = fstart + real(i-1) * df
     write (*,*) ' Freq. = ', filtfrq
!
!    length of segments = wfac * period of actual filter frequency
!    -------------------------------------------------------------
     tseg = wfac / filtfrq
!
!    ...work
!    -------
     call calc(tseg, filtfrq, iwin, dtout, tout, ampl, nsegout)
!
!    store time and amplitude (NB: due to varying window
!    width a different time vector results for each frequency)
!    ---------------------------------------------------------
     do j = 1, nsegout
        astor(j,i) = ampl(j)
        tstor(j,i) = tout(j)
     end do
  end do
!
! save results for external post-processing
! -----------------------------------------
  open (20, file = fnout, form = 'formatted', iostat = iocheck)
  if (iocheck .ne. 0 ) then
     write (*, '(//a)') ' Error - Can''t create ', trim(fnout)
     stop ' '
  end if
  write (20,'(a)')          '# ' // prgvers
  write (20,'(a9,a)')       '# File = ', trim(fnin)
  write (20,'(a11, e10.4)') '# fstart = ', fstart
  write (20,'(a10, e10.4)') '# fstop = ', filtfrq
  write (20,'(a7, e10.4)')  '# df = ', df
  write (20,'(a9, i4)')     '# NFrq = ', nfrq
  write (20,'(a9, f4.1)')   '# WFac = ', wfac
  write (20,'(a9, i2)')     '# IWin = ', iwin
  write (20,'(a10, e10.4)') '# dtout = ', dtout
  write (20,'(a38, i3)')    '# Occurrences of fstop > local f_Nyq: ', nyqerr
  write (20,'(a1)')         '#'
  write (20,'(a36)')        '#  Time       Frequency    Amplitude'
  do i = 1, nfrq
     fnow = fstart + real(i-1) * df
     do j = 1, np
        if (astor(j,i) .ne. dumval) then
           write (20,'(3(e12.6,1x))') tstor(j,i), fnow, astor(j,i)
        end if
     end do
     if (gplt .eqv. .true.) write(20,*)
  end do
  close(20)
!
! clean up
! --------
  deallocate (astor)
  deallocate (tstor)
!
! Frequency response of the harmonic filter (used for empirical
! determination of the bandwidth). Create test signals (time axis
! of input data) with varying frequency (wobble around f0). For
! each time series, the average amplitude of the harmonic filter
! (constant filter frequency f0) is saved vs. actual freq.
! ---------------------------------------------------------------
  if (bwtest .eqv. .true.) then
     filtfrq = 0.5 * (fstart + fstop)
     f0 = filtfrq
     df = (fstop - fstart) / 201.
     tseg = wfac / filtfrq
!
     open (30, file = 'bw.plt', form = 'formatted')
     write (30,*) 'plot ''-'''
!
!    loop over frequencies
!    ---------------------
     do i = -100, 100
        fnow = f0 + df * real(i)
        write (*,*) ' BWTest: Freq. = ', fnow
        arg = 2.0 * pi * fnow
        x(1:np) = sin(arg*t(1:np))
        call calc(tseg, filtfrq, iwin, dtout, tout, ampl, nsegout)
        write (30,*) fnow, sum(ampl(1:nsegout)) / real(nsegout)
     end do
     close(30)
  end if
!
  end
!
!
!--------------------------------------------------------------------------
  subroutine readdat(fnam, t, x, np)
!--------------------------------------------------------------------------
  use const
  implicit none
!
  character (len = *) :: fnam
  real, dimension(maxdim) :: t, x
  integer :: np
  integer :: i, iocheck
  character (len = 1) :: flag
!
! open input file
! ---------------
  open (10, file = fnam, form = 'formatted', status = 'old', &
       iostat = iocheck)
  if (iocheck .ne. 0 ) then
     write (*, '(//a)') ' Error - Can''t open '// trim(fnam)
     stop ' '
  end if
!
! skip header
! -----------
  do while (.true.)
     read (10, '(a1)') flag
     if (flag .ne. '#') then
        backspace (10)
        exit
     end if
  end do
!
! retrieve data
! -------------
  i = 1
  do while (.true.)
     read (10, *, iostat = iocheck) t(i), x(i)
     if (iocheck .ne. 0) exit
     i = i + 1
  end do
  close(10)
  np = i - 1
!
  end subroutine readdat
!
!
!--------------------------------------------------------------------------
  subroutine check(t, x, np)
!--------------------------------------------------------------------------
  use const
  implicit none
!
  real, dimension(maxdim) :: t, x
  integer :: np
  real    :: ave
  integer :: i, j, idx, icnt
  logical :: err_order = .false.
  logical :: err_dupl = .false.
!
! check for descending time axis
! ------------------------------
  do i = 1, np-1
     if (t(i) .gt. t(i+1)) then
        err_order = .true.
        write(*,'(1x,a,1x,e12.6)') 'Descending time axis at t =', t(i)
     end if
  end do
  if (err_order .eqv. .true.) stop
!
! check for duplicates time and replace values of time-dependent variable by their mean
! -------------------------------------------------------------------------------------
  idx = 1
  do while (.true.)
     if (t(idx+1) .eq. t(idx)) then
        err_dupl = .true.
        write(*,'(1x,a,1x,e12.6,1x,a)') 'Duplicate time at t =', t(idx), '...averaging data'
!
!       search for multiple occurrences
!       -------------------------------
        icnt = 0
        do i = 1, np-idx
           if (t(idx+i) .eq. t(idx)) icnt = icnt + 1
        end do
!
!       replace first data by mean of duplicates points
!       -----------------------------------------------
        ave = sum(x(idx:idx+icnt)) / real(icnt+1)
        x(idx) = ave
!
!       shift remaining points
!       ----------------------
        do j = 1, icnt
           do i = idx+1, np-1
                t(i) = t(i+1)
                x(i) = x(i+1)
           end do
           np  = np - 1
        end do
     end if
     idx = idx + 1
     if (idx .eq. np-1) exit
  end do
!
! save averaged data set
! ----------------------
  if (err_dupl .eqv. .true.) then
     open(10, file = "TimeSeries.avg")
     do i = 1, np
        write(10,*) t(i), x(i)
     end do
     close(10)
  end if
!
  end subroutine check
!
!
!---------------------------------------------------------------------------
  subroutine calc(tseg, filtfrq, iwin, dtout, tout, ampl, nsegout)
!---------------------------------------------------------------------------
  use const
  use timeser
!
  implicit none
!
  real :: tseg, filtfrq, dtout
  real, dimension(maxdim) :: tout, ampl
  integer :: iwin, nsegout
  real, dimension(:), allocatable :: xwk, twk
  real    :: t0, avg, told
  integer :: segcnt, npseg, start, idx
  logical :: enddat, ini
!
! loop over segments
! ------------------
  segcnt = 0
  start = 0
  enddat = .false.
  ini = .true.
  do
     if (ini .eqv. .true.) then
        start = 1
        ini = .false.
     else if (dtout .gt. 0.0) then
        told = t(start)
        do while (.true.)
           start = start + 1
           if ((t(start) .ge. told + dtout) .or. (start .eq. np)) exit
        end do
     else if (dtout .eq. 0.0) then
        start = start + 1
     end if
     segcnt = segcnt + 1
!
!    start time of the segcnt'th segment
!    -----------------------------------
     t0 = t(start)
     idx = start
!
!    find end of the segment
!    -----------------------
     do
       idx = idx + 1
       if (idx .ge. np) exit       ! separate conditions to prevent t(idx) being undefined for idx .gt. np
       if (t(idx)-t0 .ge. tseg) exit
     end do
     if (idx .ge. np) then
        enddat = .true.
        exit
     end if
!
!    points in the segment
!    ---------------------
     npseg = idx - start + 1
!
!    copy segment data into temporary workspace
!    ------------------------------------------
     allocate(xwk(npseg))
     allocate(twk(npseg))
     xwk(1:npseg) = x(start:start+npseg-1)
     twk(1:npseg) = t(start:start+npseg-1)
!
!    average observation interval for the segment -> used for output
!    ---------------------------------------------------------------
     tout(segcnt) = 0.5 * (twk(1) + twk(npseg))
!
!    subtract mean
!    -------------
     avg = sum(xwk(1:npseg)) / real(npseg)
     xwk(1:npseg) = xwk(1:npseg) - avg
!
!    taper data
!    ----------
     call taper(twk(1:npseg), xwk(1:npseg), npseg, iwin)
!
!    determine amplitude/phase of the best fit sinusoidal curve
!    ----------------------------------------------------------
     call sinfit (filtfrq, xwk, twk, npseg, ampl(segcnt))
     deallocate(xwk)
     deallocate(twk)
     if (enddat .eqv. .true.) exit
  end do
  nsegout = segcnt - 1
!
  end subroutine calc
!
!
!--------------------------------------------------------------------------
  subroutine taper(t, x, np, iwin)
!--------------------------------------------------------------------------
! apply taper to input data
!
! window type (iwin)  0: Rectangular
!                     1: Welch 1
!                     2: Hanning
!                     3: Parzen (Triangular)
!                     4: Blackman-Harris 3-Term
!--------------------------------------------------------------------------
  use const
!
  implicit none
!
  integer, intent(in) :: np, iwin
  real, intent(in), dimension(np) :: t
  real, intent(inout), dimension(np) :: x
!
  real, dimension(maxdim) :: ww
  real :: tlen, jeff, scal, fac1, fac2, fac3, fac4, sumw, rnp
  integer :: i
!
! useful factor for various windows
! ---------------------------------
  rnp = real(np)
  fac1 = (rnp / 2.0 ) - 0.5
  fac2 = 1.0 / ((rnp / 2.0 ) + 0.5)
  fac3 = rnp - 1.0
  fac4 = twopi /(rnp - 1.0)
  tlen = t(np) - t(1)
  sumw = 0.0
  do i = 1, np
      jeff = rnp * (t(i)-t(1)) / tlen
      select case (iwin)
         case (0)                                            ! rectangle
            ww(i) = 1.0
         case (1)                                            ! welch I
            ww(i) = 1.0 - ((jeff - fac1) * fac2)**2.0
         case (2)                                            ! hanning
            ww(i) = 0.5 * (1.0 - cos(twopi * jeff / fac3))
         case (3)                                            ! triangular
            ww(i) = 1.0 - abs((jeff - fac1) * fac2)
         case (4)                                            ! blackman-harris
            ww(i) = 0.4243801 - 0.4973406 * cos(fac4 * jeff) &
                    + 0.0782793 * cos(fac4 * 2.0 * jeff)
      end select
      sumw = sumw + ww(i)
  end do
!
! set sum of window weights = 1
! -----------------------------
  scal = rnp / sumw
  do i = 1, np
     ww(i) = ww(i) * scal
  end do
!
! apply taper
! -----------
  x(1:np) = ww(1:np) * x(1:np)
!
  end subroutine taper
!
!
!---------------------------------------------------------------------------
  subroutine sinfit (filtfrq, xwk, twk, np, ampl)
!---------------------------------------------------------------------------
  use const
!
  implicit none
!
  integer :: np
  real, dimension(np) :: xwk, twk
  real :: ampl, filtfrq
  real :: arg, denom, tmp, sum, corr, d0, d1, d2, c1, c2, a0, rnp, &
          a1, a2, scos, ssin, scos2, ssin2, sprod, sxcos, sxsin, s, c
  integer :: i
!
! make sum of data equal to zero (eq. 3 in FM81)
! --------------------------------------
  corr = sum(xwk(1:np)) / real(np)
  xwk(1:np) = xwk(1:np) - corr
!
! trigonometric sums
! ------------------
  scos = 0.0
  ssin = 0.0
  scos2 = 0.0
  ssin2 = 0.0
  sxcos = 0.0
  sxsin = 0.0
  sprod = 0.0
  arg = 2.0 * pi * filtfrq
  do i = 1, np
     s = sin(arg*twk(i))
     c = cos(arg*twk(i))
     scos = scos + c
     ssin = ssin + s
     scos2 = scos2 + c*c
     ssin2 = ssin2 + s*s
     sprod = sprod + s*c
     sxcos = sxcos + xwk(i) * c
     sxsin = sxsin + xwk(i) * s
  end do
!
! determine coefficients (eqs. 13 & 15 in FM81; w_i = 0 for all i)
! ----------------------------------------------------------------
  rnp = real(np)
  a0 = sqrt(1.0 / rnp)
  denom = abs(scos2 - scos*scos/rnp)
  a1 = sqrt(1.0/denom)
  tmp = sprod - a0*a0*ssin*scos
  denom = abs(ssin2 - ssin*ssin/rnp - a1*a1*tmp*tmp)
  a2 = sqrt(1.0/denom)
  c1 = a1*sxcos
  c2 = a2*sxsin - a1*a2*c1*tmp
!
! filter coefficients (p. 623 in FM81)
! ------------------------------------
  d2 = a2*c2
  tmp = a0*a0*ssin*scos - sprod
  d1 = a1*c1 + d2*a1*a1*tmp
  d0 = -a0*a0 * (d1*scos + d2*ssin)
!
! amplitude and phase of filter function
! --------------------------------------
  ampl = abs(d0) + sqrt(d1*d1 + d2*d2)
!
  end subroutine sinfit
